'''
 Bjoern Annighoefer 2022
 
'''

from ..error import EOQ_ERROR_DOES_NOT_EXIST

class History:
    ''' A class to store the history of a compound command
    
    '''
    
    def __init__(self):
        self.values = []
        self.valueNames = {}
        self.n = 0
        self.nMute = 0
        
    def AddValue(self,v,mute:bool,name:str=None):
        self.values.append(v)
        if(name):
            self.valueNames[name] = (self.n,mute,self.nMute)
        self.n += 1
        if(not mute): self.nMute += 1 #build a second index only including muted values
        
    def GetValueByIndex(self,i:int):
        try: 
            return self.values[i]
        except KeyError:
            raise EOQ_ERROR_DOES_NOT_EXIST('History index %d is out of range'%(i))
    
    def GetValueByName(self,n:str):
        try: 
            return self.values[self.valueNames[n][0]]
        except KeyError:
            raise EOQ_ERROR_DOES_NOT_EXIST('History name %s is unknown '%(n))
    
    def GetValueNamesAndIndicies(self):
        valueNames = []
        valueIndicies = []
        for (k,v) in self.valueNames.items():
            if(not v[1]): #is not muted? 
                valueNames.append(k)
                valueIndicies.append(v[2]) #only append the muted index, which is correct for the reduced history
        return (valueNames,valueIndicies)
            


'''
 2019 Bjoern Annighoefer
'''

from ..domain import Domain
from ..config import Config
from ..logger import GetLoggerInstance
from ..serializer import CreateSerializer
#type checking
from typing import Tuple

class SerialDomain(Domain):
    '''A domain that supports processing serialized commands directly, 
    which is often better performing in multi-computer environments
    '''
    def __init__(self, config:Config):
        super().__init__()
        self.config = config
        self.logger = GetLoggerInstance(config)
        self.remoteFrmSerializer = CreateSerializer(config.remoteFrmSerializer)
        self.remoteCmdSerializer = CreateSerializer(config.remoteCmdSerializer)
        
    def SerRawDo(self, cmdStr:str, sessionId:str=None, serType:str=None, readOnly:bool=False)->Tuple[str,str]:
        '''Direct processing of serialized commands
        Args:
        - cmdStr: The serialized command
        - sessionId: the session under which this command is processed
        - serType: [3 letter str] the name of the serializer
        
        Returns:
        - resStr: The serialized results
        - serType: [3 letter str] The serializer name used for the results
        '''
        raise NotImplemented()
            
    
    def Close(self):
        pass #is to be called if the domain closes down
    
                    
    
        
        


'''
 Bjoern Annighoefer 2022
 
'''

from ..value import LST
from ..query import Obj

from datetime import datetime

from typing import Tuple, Dict

class Session:
    ''' The class to store the session information when interacting with the domain
    
    '''
    def __init__(self, sessionId:str, sessionIdPublic:str, sessionIdSecret:str, sessionNumber:int):
        self.sessionId = sessionId
        self.sessionIdPublic = sessionIdPublic
        self.sessionIdSecret = sessionIdSecret
        self.sessionNumber = sessionNumber
        self.user = None
        self.begin = datetime.now()
        self.isTemp = False
        #initialize event table
        self.observedEventTypes:Dict[str,bool] = {} #the event types to observe -> bool
        self.superobservedEventTypes:Dict[str,bool] = {} #the event types to observe -> bool
        self.observedElements:Dict[Obj,bool] = {} #stores the elements known by the session
        self.watchedQueries:Dict[int,bool] = {} #stores the keys to watched expressions
        self.observedMessageKeys:Dict[str,bool] = {} #stores all message keys to be observed
        
    ### REGULAR OBSERVATION ###
            
    def ObserveEvent(self,evtType:str)->bool:
        '''Starts listening to this type of events
        Returns: is this a new observation?
        '''
        if(evtType in self.observedEventTypes):
            return False #is already observed
        else:
            self.observedEventTypes[evtType] = True
            return True
        
    def UnobserveEvent(self, evtType:str)->bool:           
        if evtType in self.observedEventTypes:
            del self.observedEventTypes[evtType]
            return True
        else:
            return False
            
    def IsEventObserved(self, evtType:str)->bool:
        if evtType in self.observedEventTypes:
            return True
        else:
            return False
        
    ### SUPER OBSERVATION ###
            
    def SuperobserveEvent(self,evtType:str)->bool:
        '''Starts listening to this type of events
        
        '''
        if(evtType in self.superobservedEventTypes):
            return False #is already observed
        else:
            self.superobservedEventTypes[evtType] = True
            return True
        
    def UnSuperobserveEvent(self, evtType:str)->bool:           
        if evtType in self.superobservedEventTypes:
            del self.superobservedEventTypes[evtType]
            return True
        else:
            return False
            
    def IsEventSuperobserved(self, evtType:str)->bool:
        if evtType in self.superobservedEventTypes:
            return True
        else:
            return False
        
    ### ELEMENT OBSERVATION ###
    
    def ObserveElement(self, elem:Obj)->bool:
        '''Add one element to the observed elements 
        and return if this was new.
        '''
        if(elem in self.observedElements):
            return False
        else:
            self.observedElements[elem] = True 
            return True
        
    def UnobserveElement(self, elem:Obj)->bool:
        if(elem in self.observedElements):
            del self.observedElements[elem] 
            return True
        else:
            return False
        
    def ObserveElements(self, elements:LST)->Tuple[bool,LST]:
        '''Add multiple elements to the observed elements 
        and return if at least one was new 
        and return a list of all new elements.
        '''
        newElements = LST([])
        isChanged = False
        for e in elements:
            if(e not in self.observedElements):
                self.observedElements[e] = True
                newElements.append(e)
                isChanged = True
        return isChanged, newElements
    
    def UnobserveElements(self, elements:LST)->Tuple[bool,LST]:
        '''Remove multiple elements from the observed elements 
        and return if at least one was removed
        and return a list of all removed elements.
        '''
        removedElements = LST([])
        isChanged = False
        for e in elements:
            if(e in self.observedElements):
                del self.observedElements[e]
                removedElements.append(e)
                isChanged = True
        return isChanged, removedElements
            
    def IsElementObserved(self, elem:Obj)->bool:
        return (elem in self.observedElements)
    
    def UnobserveAllElements(self)->Tuple[bool,LST]:
        '''Remove all elements from the observed elements 
        and return if at least one was removed
        and return a list of all removed elements.
        '''
        removedElements = LST([e for e in self.observedElements.keys()])
        self.observedElements.clear()
        isChanged = 0<len(removedElements)
        return isChanged, removedElements
    
    ### QUERY OBSERVATION ###
    
    def ObserveQuery(self, watchId:int)->bool:
        if(watchId in self.watchedQueries):
            return False
        else:
            self.watchedQueries[watchId] = True
            return True
        
    def UnobserveQuery(self, watchId:int)->None:
        if(watchId in self.watchedQueries):
            del self.watchedQueries[watchId]
            return True
        else:
            return False
            
    def ClearObservedQueries(self)->None:
        self.watchedQueries = {}
            
    def IsQueryObserved(self, watchId:int)->bool:
        return (watchId in self.watchedQueries)
    
    def GetWatchedQueryIds(self)->list:
        return [wid for wid in self.watchedQueries.keys()]
    
    ### MEASSAGE OBSERVATION ###
    
    def ObserveMessage(self, msgKey:str)->None:
        if(msgKey in self.observedMessageKeys):
            return False
        else:
            self.observedMessageKeys[msgKey] = True
            return True
        
    def UnobserveMessage(self, msgKey:str)->None:
        if(msgKey in self.observedMessageKeys):
            del self.observedMessageKeys[msgKey]
            return True
        else:
            return False
            
    def IsMessageObserved(self, msgKey:str)->bool:
        return (msgKey in self.observedMessageKeys) 
        
    
    
    


'''
 2019 Bjoern Annighoefer
'''

from .history import History
from .util import ApplyToAllElements,ApplyToAllElementsInA,ApplyToAllElementsInB,ApplyToAllListsOfElementsInA,ApplyToSimilarElementStrutures,ApplyToSimilarListsOfObjects,IsList,IsListOfObjects,TRM,Determinate,IsNoList

from ..config import Config
from ..logger import GetLoggerInstance
from ..concepts import CONCEPTS, MXMDB, MXELEMENT, MXCONSTRAINT,\
                       M2MODEL, M2PACKAGE, M2ENUM,M2OPTIONOFENUM, M2CLASS, M2ATTRIBUTE, M2ASSOCIATION, M2COMPOSITION, M2INHERITANCE,\
                       M1MODEL, M1OBJECT, M1COMPOSITION, M1ATTRIBUTE, M1ASSOCIATION
from ..value import VAL, PRM, BOL, U32, U64, I32, I64, F32, F64, STR, DAT, NON, LST, QRY, ValidateVal
from ..query import SEG_TYPES, Seg, Qry, Obj, MET_MODES, IDX_MODES, STF_MODES
from ..error import EOQ_ERROR, EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_INVALID_VALUE, EOQ_ERROR_DOES_NOT_EXIST, EOQ_ERROR_UNSUPPORTED, EOQ_ERROR_UNKNOWN, EOQ_ERROR_RUNTIME
from ..serializer import Serializer, TextSerializer
from ..mdb import Mdb
from ..util.benchmark import Benchmark

import re

# constants
CONTEXT_LESS_SEGMENTS = {
    SEG_TYPES.OBJ : True,
    SEG_TYPES.HIS : True
    }


# Special type handling regex
class RGX(STR):
    def __init__(self, value:re.Pattern):
        VAL.__init__(self,'RGX')
        self.v = value
    def __repr__(self):
        return str(self.v)
    def Type(self):
        return STR #regex are handled like strings

#terminate operation
def TvX(a,b):
    return a #do nothing


class QryRunner:
    '''QueryRunner
        
        Evaluates a query using a mdb. Only the EvalOnContextAndScope function should be called externally
        
    '''
    def __init__(self, mdb:Mdb, config:Config):
        self.mdb:Mdb = mdb
        self.mdbObj:Obj = None
        self.config:Config = config
        #internals
        self.qrySerializer:Serializer = TextSerializer()
        self.segmentEvaluators = {}
        self.logger = GetLoggerInstance(config)
        #benchmark
        if self.config.enableStatistics:
            self.benchmark = Benchmark()
            allQryTypes = [getattr(SEG_TYPES,k) for k in SEG_TYPES.__dict__ if not k.startswith('_')]
            for t in allQryTypes:
                self.benchmark.InitMessure(t)
        #Segment operators
        self.segmentEvaluators = self.__InitQrySegHandlers()
        #META operators
        self.metEvaluators = self.__InitMetOperationHandlers()
        # init numerical and boolean operation handlers
        self.equEvaluators = self.__InitEquOperationHandlers()
        self.neqEvaluators = self.__InitNeqOperationHandlers()
        self.greEvaluators = self.__InitGreOperationHandlers()
        self.lesEvaluators = self.__InitLesOperationHandlers()
        self.rgxEvaluators = self.__InitRgxOperationHandlers()
        self.addEvaluators = self.__InitAddOperationHandlers()
        self.subEvaluators = self.__InitSubOperationHandlers()
        self.mulEvaluators = self.__InitMulOperationHandlers()
        self.divEvaluators = self.__InitDivOperationHandlers()
        # init list length handler
        self.lenFunctor = lambda a, b: U64(len(a))
        # init list operation handlers
        #cross product operation
        def cspUni(a,b):
            res = LST()
            for e1 in a:
                for e2 in b:
                    res.append(LST([e1,e2]))
            return res
        self.cspEvaluator = cspUni
        #intersection operation
        def itsUni(a,b):
            res = LST()
            #add common elements
            for e1 in a:
                if(e1 in b):
                    res.append(e1)
            return res
        self.itsEvaluator = itsUni
        #set difference operation
        def difUni(a,b):
            res = LST()
            #add common elements
            for e1 in a:
                if(e1 not in b):
                    res.append(e1)
            return res
        self.difEvaluator = difUni
        #union operations
        def uniUni(a,b):
            res = LST()
            #add all unique elements of a
            for e in a:
                if(e in res):
                    continue
                res.append(e)
            #add all unique elements of b
            for e in b:
                if(e in res):
                    continue
                res.append(e)
            return res
        self.uniEvaluator = uniUni
        #concatenate operation
        def conUni(a,b):
            res = []
            res.extend(a)
            res.extend(b)
            return LST(res)
        self.conEvaluator = conUni

        
    def Eval(self, val:VAL, history:History=None) -> VAL:
        context = NON()
        #get an initial context if the first segment requires that 
        if(isinstance(val,Qry) and 0<len(val.v) and val.v[0].seg not in CONTEXT_LESS_SEGMENTS):
            if(None==self.mdbObj):
                # init mdb obj to prevent that this must be re-read
                self.mdbObj = self.mdb.Read(NON(),STR(MXMDB.MDB)) #read non is MXDB
            modelroot = self.mdb.Read(self.mdbObj,STR(MXMDB.ROOT)) #read non is MXDB
            context = self.mdbObj if modelroot.IsNone() else modelroot
        return self.EvalOnContextAndScope(context,val,context,history)
        
    def EvalOnContextAndScope(self, context:VAL, val:VAL, scope:VAL, history:History) -> VAL:
        if(isinstance(val,Qry)):
            return self.EvalQry(context, val.v, scope, history)
        elif(IsList(val)):
            return self.EvalArr(context,LST([val]),scope,history)
        else:
            return val #its a primitive value, do nothing 
    
    '''
        SEGMENT EVALUATORS
    '''
    def EvalSeg(self, context:VAL,seg : Seg, scope:VAL, history:History) -> VAL:
        res = NON()
        segType = seg.seg
        try:
            evalFunction = self.segmentEvaluators[segType] 
        except KeyError as e:
            raise EOQ_ERROR_UNSUPPORTED("Unknown segment type: %s: %s"%(segType,str(e)))
        if self.config.enableStatistics: self.benchmark.Start()
        try:
            v = seg.v 
            res = evalFunction(context,v,scope,history)
        except Exception as e:
            if self.config.enableStatistics: self.benchmark.Stop(segType)
            raise e         
        if self.config.enableStatistics: self.benchmark.Stop(segType)
        return res
   
        
    def EvalQry(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        currentContext = scope #each subquery restarts from the current scope
        #newScope = context
        n = 1
        for seg in args:
            if(isinstance(currentContext,TRM)):
                break;
            try:
                currentContext = self.EvalSeg(currentContext,seg,scope,history)
            except EOQ_ERROR as e:
                e.msg = "Segment %d (%s): "%(n,seg.seg) + e.msg
                raise e
            except Exception as e:
                raise EOQ_ERROR_UNKNOWN("Segment %d (%s): %s"%(n,seg.seg,str(e)))
            n+=1
        res = Determinate(currentContext)
        return res
    
    def EvalObj(self, context:VAL, args:list, scope:VAL, history:History) -> VAL :
        objId = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[PRM],'element ID',exact=False)
        #TODO: speed up local processing, by not creating an new object seg element, but by attaching the old one.
        return Obj(objId) # because the argument is unpacked before
    
    def EvalOni(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = NON()
        nameOrId = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'nameOrId')
        ValidateVal(context,[Obj],'context',False)
        elements = self.mdb.FindElementByIdOrName(nameOrId,context)
        n = len(elements)
        if(0==n):
            raise EOQ_ERROR_DOES_NOT_EXIST("%s is no element name or ID."%(nameOrId))
        if(n>1):
            raise EOQ_ERROR_INVALID_VALUE("%s is not unique"%(nameOrId))
        else:
            res = elements[0]         
        return res
    
    def EvalHis(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        ref = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[I64,STR],'reference')
        if(None != history):
            if(isinstance(ref,I64)):
                return history.GetValueByIndex(ref.GetVal())
            else:
                return history.GetValueByName(ref.GetVal())
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST('No history available. Used his outside a compound command?')
        
    def EvalPth(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        name = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'name')
        pathFunctor = lambda o: self.mdb.Read(o,name)
        res = ApplyToAllElements(context, pathFunctor)
        return res
    
    def EvalCls(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = NON()
        name = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'name')
        classes = self.mdb.FindElementByIdOrName(name,NON(),STR(CONCEPTS.M2CLASS))
        nClasses = len(classes)
        if(nClasses>0):
            if(nClasses>1):
                self.logger.Warn('Class name %s is not unique. Use class ID instead.'%(name))
            res = LST()
            clsFunctor = lambda a,b: self.mdb.Read(b,STR(M2CLASS.INCARNATIONS),a)
            for c in classes:
                res += LST(ApplyToAllElementsInA(context,c,clsFunctor))
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST("No known class with name %s."%(name))
        return res

    def EvalIno(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = NON()
        name = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'name')
        classes = self.mdb.FindElementByIdOrName(name,NON(),STR(CONCEPTS.M2CLASS))
        nClasses = len(classes)
        if(nClasses>0):
            if(nClasses>1):
                self.logger.Warn('Class name %s is not unique. Use class ID instead.'%(name))
            res = LST()
            inoFunctor = lambda a,b: self.mdb.Read(b,STR(M2CLASS.INSTANCES),a)
            for c in classes:
                res += LST(ApplyToAllElementsInA(context,c,inoFunctor))
        else:
            raise EOQ_ERROR_DOES_NOT_EXIST("No known class with name %s."%(name))
        return res

    
    def EvalNot(self, context:VAL, args:VAL, scope:VAL, history:History) -> VAL:
        notFunctor = lambda o: BOL(False) if o.IsTrue() else BOL(True)
        res = ApplyToAllElements(context, notFunctor)
        return res
    
    def EvalTrm(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        #Define select functors
        def TrmOperator(a,b,c):
            res = None
            if(isinstance(a,TRM)):
                res = a
            elif(b):
                res = TRM(c)
            else:
                res = a
            return res
        
        def TrmElemVsElemFunc(a,b,c):
            return LST([TrmOperator(a[i],b[i],c) for i in range(len(b))])

        def TrmElemVsStructFunc(a,b,c):
            raise EOQ_ERROR_RUNTIME("Error applying termination: Argument of termination condition must be of lower depth than the context, but got %s{%s,%s}="%(a,b,c))
        
        def TrmStructVsElemFunc(a,b,c): 
            return LST([TrmOperator(a[i],b[i],c) for i in range(len(b))])
        #Begin of function
        condquery = ValidateVal(args[0],[VAL],'condition',False)
        if(NON()==condquery): #special default case
            condquery = Qry().Equ(NON())
        condition = self.EvalOnContextAndScope(context,condquery,context,history)
        default = self.EvalOnContextAndScope(context,args[1],context,history)
        res = ApplyToSimilarListsOfObjects(LST([context]),LST([condition]),TrmElemVsElemFunc,TrmElemVsStructFunc,TrmStructVsElemFunc,default)
        return res[0] #return the first element because context and condition were listified above
    
    def EvalQrf(self, context:VAL, args:VAL, scope:VAL, history:History) -> VAL:
               
        def qrfFunctor(a:Obj):
            if(isinstance(a,Obj)):
                #reverse the path through concepts to this element
                pathFromParent = []
                currentElement = a
                currentConcept = self.mdb.Read(currentElement,STR(MXELEMENT.CONCEPT)).GetVal()
                while(currentConcept != CONCEPTS.MXMDB):
                    parent = None
                    feature = None
                    index = None
                    # MX layer
                    if(currentConcept==CONCEPTS.MXCONSTRAINT):
                        parent = self.mdb.Read(currentElement,STR(MXCONSTRAINT.ELEMENT))
                        feature = STR(MXELEMENT.CONSTRAINTS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    # M2 layer
                    elif(currentConcept==CONCEPTS.M2MODEL):
                        (parent,perr) = self.mdb.Create(STR(CONCEPTS.MXMDB))
                        feature = STR(MXMDB.M2MODELS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2PACKAGE):
                        parent = self.mdb.Read(currentElement,STR(M2PACKAGE.SUPERPACKAGE))
                        feature = STR(M2PACKAGE.SUBPACKAGES)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2ENUM):
                        parent = self.mdb.Read(currentElement,STR(M2ENUM.M2PACKAGE))
                        feature = STR(M2PACKAGE.ENUMS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2OPTIONOFENUM):
                        parent = self.mdb.Read(currentElement,STR(M2OPTIONOFENUM.ENUM))
                        feature = STR(M2ENUM.OPTIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2CLASS):
                        parent = self.mdb.Read(currentElement,STR(M2CLASS.M2PACKAGE))
                        feature = STR(M2PACKAGE.CLASSES)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2ATTRIBUTE):
                        parent = self.mdb.Read(currentElement,STR(M2ATTRIBUTE.SRCCLASS))
                        feature = STR(M2CLASS.MYATTRIBUTES)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2ASSOCIATION):
                        parent = self.mdb.Read(currentElement,STR(M2ASSOCIATION.SRCCLASS))
                        feature = STR(M2CLASS.MYSRCASSOCIATIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2COMPOSITION):
                        parent = self.mdb.Read(currentElement,STR(M2COMPOSITION.PARENTCLASS))
                        feature = STR(M2CLASS.MYPARENTCOMPOSITIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M2INHERITANCE):
                        parent = self.mdb.Read(currentElement,STR(M2INHERITANCE.SUBCLASS))
                        feature = STR(M2CLASS.SPECIALIZATIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    # M1 layer
                    elif(currentConcept==CONCEPTS.M1MODEL):
                        (parent,perr) = self.mdb.Create(STR(CONCEPTS.MXMDB))
                        feature = STR(MXMDB.M1MODELS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M1OBJECT):
                        parent = self.mdb.Read(currentElement,STR(M1OBJECT.MODEL))
                        feature = STR(M1MODEL.OBJECTS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M1ATTRIBUTE):
                        parent = self.mdb.Read(currentElement,STR(M1ATTRIBUTE.OBJECT))
                        feature = STR(M1OBJECT.ATTRIBUTES)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M1ASSOCIATION):
                        parent = self.mdb.Read(currentElement,STR(M1ASSOCIATION.SRC))
                        feature = STR(M1OBJECT.SRCASSOCIATIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    elif(currentConcept==CONCEPTS.M1COMPOSITION):
                        parent = self.mdb.Read(currentElement,STR(M1COMPOSITION.PARENT))
                        feature = STR(M1OBJECT.PARENTCOMPOSITIONS)
                        index = I64(self.mdb.Read(parent,feature).index(currentElement)) #low performance
                    #prepare next round
                    pathFromParent.append((parent,feature,index))
                    currentElement = parent
                    currentConcept = self.mdb.Read(currentElement,STR(MXELEMENT.CONCEPT)).GetVal()
                #build the path in reverse order
                val = Qry()
                for segment in reversed(pathFromParent):
                    f = segment[1]
                    i = segment[2]
                    if(None == i):
                        val.Pth(f)
                    else:
                        val.Pth(f).Idx(i)
            else:
                EOQ_ERROR_INVALID_TYPE("Qrf only requires context of type OBJ.")
            return val
            
        res = ApplyToAllElements(context, qrfFunctor)
        return res
    
    def EvalStf(self, context:VAL, args:VAL, scope:VAL, history:History) -> VAL:
        mode = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'mode').GetVal()
        res = NON()
        if(STF_MODES.ELM==mode):
            stfFunctor = lambda o: STR(self.qrySerializer.SerVal(o))
            res = ApplyToAllElements(context, stfFunctor)
        elif(STF_MODES.LST==mode):
            if(LST!=type(context)):
                raise EOQ_ERROR_INVALID_TYPE("Only LST can be input for stringify in mode LST.")
            res = LST([STR(self.qrySerializer.SerVal(o)) for o in context])
        elif(STF_MODES.FUL==mode):
            res = STR(self.qrySerializer.Ser(context))
        else:
            raise EOQ_ERROR_INVALID_VALUE("Unknown mode: %s"%(mode)) 
        return res
    
    def EvalSlf(self, context:VAL, args:VAL, scope:VAL, history:History) -> VAL:
        return context
    
    def EvalTry(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        def tryFunctor(a,b):
            res = NON()
            query = b[0]
            default = b[1]
            history = b[2]
            try:
                res = self.EvalOnContextAndScope(a,query,a,history)
            except:
                res = self.EvalOnContextAndScope(a,default,a,history)
            return res
        
        query = ValidateVal(args[0],[VAL],'tryStatement',False) #do not evaluate here but element-wise inside try
        default = ValidateVal(args[1],[VAL],'exceptStatement',False)

        res = ApplyToAllElementsInA(context,(query,default,history),tryFunctor)
        
        return res
    
    def EvalIdx(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = NON()
        if(not IsList(context)):
            raise EOQ_ERROR_INVALID_TYPE("Can only select from LST but got: %s"%(context))
        nArgs = len(args)
        if(1 == nArgs):
            n = self.EvalOnContextAndScope(context,args[0],context,history)
            if(isinstance(n, I64)):
                try:
                    idxFunctor = lambda a,b: a[b.GetVal()]
                    res = ApplyToAllListsOfElementsInA(context,n,idxFunctor)
                except IndexError:
                    raise EOQ_ERROR_DOES_NOT_EXIST('Index %d is out of bounds.'%(n.GetVal()))
            elif(isinstance(n, STR)):
                if(STR(IDX_MODES.ASC)==n): #sort ascending
                    ascFunctor = lambda a,b: LST(sorted(a))
                    res = ApplyToAllListsOfElementsInA(context,None,ascFunctor)
                elif(STR(IDX_MODES.DSC)==n): #sort descending
                    dscFunctor = lambda a,b: LST(sorted(a,reverse=True))
                    res = ApplyToAllListsOfElementsInA(context,None,dscFunctor)
                elif(STR(IDX_MODES.FLT)==n): #flatten
                    if(IsList(context)):
                        res = LST()
                        self.__Flatten(context,res)
                    else:
                        res = context
                elif(STR(IDX_MODES.LEN)==n): #calc size
                    res = ApplyToAllListsOfElementsInA(context,NON(),self.lenFunctor)
                else:
                    raise EOQ_ERROR_UNSUPPORTED("Unknown keyword: %s"%(n))
            else:
                raise EOQ_ERROR_INVALID_TYPE("Argument must be STR or I64, but got %s."%(type(n).__name__))
        elif(3 == nArgs):
            n = [self.EvalOnContextAndScope(context,args[0],context,history),
                 self.EvalOnContextAndScope(context,args[1],context,history),
                 self.EvalOnContextAndScope(context,args[2],context,history)]
            if(isinstance(n[0], I64) and isinstance(n[1], I64) and isinstance(n[2], I64)):
                rngFunctor = lambda a,b: a[b[0].GetVal():b[1].GetVal():b[2].GetVal()]
                try:
                    res = ApplyToAllListsOfElementsInA(context,n,rngFunctor)
                except IndexError:
                    raise EOQ_ERROR_DOES_NOT_EXIST('Index range %s is out of bounds.'%(n))
            else:
                raise EOQ_ERROR_INVALID_TYPE("Argument must be I64[3] but got: %s"%(type(n).__name__))
        else:
            raise EOQ_ERROR_INVALID_TYPE("Argument must be I64[3] or STR but got: %s" % (args))
        return res
    
    def EvalSel(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = LST
        #Define select functors
        def SelListVsListFunc(a,b,c):
            return LST([a[i] for i in range(len(b)) if b[i].IsTrue()])
        def SelListVsStructFunc(a,b,c):
            raise EOQ_ERROR_INVALID_VALUE("Error applying selector: Argument of selector must be of lower depth than the context, but got %s{%s}"%(a,b))
        def SelStructVsListFunc(a,b,c): 
            return LST([a[i] for i in range(len(b)) if b[i].IsTrue()]) # is the same since 
        # Input check 
        if(IsNoList(context)): 
            raise EOQ_ERROR_INVALID_VALUE("Select only works on lists or lists of list, but got %s"%(str(context)))
        #Start Select evaluation        
        # selector changes the context
        if(0==len(context)):
            res = LST() #The result of an empty array is always an empty array. This saves time and prevents wrong interpretations of select queries that reduce the array length, e.g. any
        else:
            select = self.EvalOnContextAndScope(context,args[0],context,history)
            res = ApplyToSimilarListsOfObjects(context,select,SelListVsListFunc,SelListVsStructFunc,SelStructVsListFunc)
        return res
    
    def EvalArr(self, context:VAL, args:list, scope:VAL, history:History) -> VAL: 
        res = LST([self.EvalOnContextAndScope(context,a,context,history) for a in args[0]])
        return res
    
    def EvalZip(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        def ZipListVsListFunc(a,b,c):
            return a+LST([b])
        def ZipListVsStructFunc(a,b,c):
            return a+LST([b])
        def ZipStructVsListFunc(a,b,c):
            return LST([ApplyToAllListsOfElementsInA(a[i],b[i],lambda a,b: a+LST([b])) for i in range(len(b))])
        
        #prepare the results structure according to the context
        res = ApplyToAllElementsInA(context,NON(),lambda a,b: LST([]))
        #get the individual results
        if(0 < len(res)): #only add elements if the list is not emptpy, because the following code missbehaves for empty list inputs
            for a in args[0]:
                ar = self.EvalOnContextAndScope(context,a,context,history)
                #merge the individual result in the result structure derived from the context
                res = ApplyToSimilarListsOfObjects(res,ar,ZipListVsListFunc,ZipListVsStructFunc,ZipStructVsListFunc) 
        return res
    
    
    def EvalAny(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        #local functor
        def anyFunctor(a,b):
            if(IsList(b)):
                for e in b:
                    if(e in a):
                        return BOL(True)
            else:
                return BOL(b in a)
            return BOL(False)
        
        #method start
        if(not IsList(context)):
            raise EOQ_ERROR_INVALID_TYPE("Can only apply ANY for LST but got: %s"%(context))
        select = self.EvalOnContextAndScope(context,args[0],context,history)
        if(IsList(select) and not IsListOfObjects(select)):
            raise EOQ_ERROR_INVALID_TYPE("Argument must be VAL or a LST of VAL but got: %s"%(select))
        res = ApplyToAllListsOfElementsInA(context,select,anyFunctor)
        return res
    
    def EvalAll(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        #local functor
        def allFunctor(a,b):
            foundMembers = 0
            if(IsList(b)):
                for e in b:
                    if(e in a):
                        foundMembers +=1
                return BOL(len(b)==foundMembers)
            else:
                return BOL(b in a)
            return BOL(False)
        
        #method start
        if(not IsList(context)):
            raise EOQ_ERROR_INVALID_TYPE("Can only apply ALL for LST but got: %s"%(context))
        select = self.EvalOnContextAndScope(context,args[0],context,history)
        if(IsList(select) and not IsListOfObjects(select)):
            raise EOQ_ERROR_INVALID_TYPE("Argument must be VAL or a LST of VAL but got: %s"%(select))
        res = ApplyToAllListsOfElementsInA(context,select,allFunctor)
        return res
    
    def EvalMet(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = None
        try: 
            name = ValidateVal(args[0],[STR],'name').GetVal()
            metaArgs = ValidateVal(args[1],[LST],'args')
            metEvaluator = self.metEvaluators[name]
            res = metEvaluator(context,metaArgs,scope,history)
        except KeyError as e:
            raise EOQ_ERROR_UNSUPPORTED("Unknown META segment type: %s: %s"%(name,str(e)))
        return res
    
    '''
        META EVALATORS
    '''
   
    def EvalMetCls(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.__DeprecatedWarning("@%s is deprecated, please use %s instead."%(MET_MODES.CLS,M1OBJECT.M2CLASS))
        clsFunctor = lambda a,b: self.mdb.Read(a,STR(M1OBJECT.M2CLASS))
        res = ApplyToAllElementsInA(context,NON(),clsFunctor)
        return res
    
    def EvalMetCln(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.__DeprecatedWarning("DEPRECATED: @%s is deprecated, please use %s/%s instead."%(MET_MODES.CLN,M1OBJECT.M2CLASS,M2CLASS.NAME))
        clnFunctor = lambda a,b: self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.NAME))
        res = ApplyToAllElementsInA(context,NON(),clnFunctor)
        return res
    
    def EvalMetLen(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        if(not IsList(context)):
            raise EOQ_ERROR_INVALID_TYPE("SIZE only works for lists, but got: %s" %(context))
        res = ApplyToAllListsOfElementsInA(context,NON(),self.lenFunctor)
        return res
    
    def EvalMetPar(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        #self.__DeprecatedWarning("DEPRECATED: @%s is deprecated, please use %s/%s instead."%(MET_MODES.PAR,M1OBJECT.CHILDCOMPOSITION,M1COMPOSITION.PARENT))
        def parFunctor(a,b):
            parent = NON()
            parentCompo = self.mdb.Read(a,STR(M1OBJECT.CHILDCOMPOSITION))
            if(parentCompo):
                parent = self.mdb.Read(parentCompo,STR(M1COMPOSITION.PARENT))
            return parent
        res = ApplyToAllElementsInA(context,NON(),parFunctor)
        return res
    
    def EvalMetAlp(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.__DeprecatedWarning("DEPRECATED: @%s is deprecated, please use %s/%s instead."%(MET_MODES.ALP,M1OBJECT.CHILDCOMPOSITION,M1COMPOSITION.PARENT))
        def alpFunctor(a,b): 
            allParents = []
            parentCompo = self.mdb.Read(a,STR(M1OBJECT.CHILDCOMPOSITION))
            while parentCompo:
                parent = self.mdb.Read(parentCompo,STR(M1COMPOSITION.PARENT))
                allParents.append(parent)
                parentCompo = self.mdb.Read(parent,STR(M1OBJECT.CHILDCOMPOSITION))
            return LST(reversed(allParents))

        res = ApplyToAllElementsInA(context,NON(),alpFunctor)
        return res
    
    def EvalMetAso(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        self.__DeprecatedWarning("DEPRECATED: @%s is deprecated, please use %s/%s instead."%(MET_MODES.ASO,M1OBJECT.DSTASSOCIATIONS,M1ASSOCIATION.SRC))
        def asoFunctor(a,b): 
            return LST([self.mdb.Read(x,STR(M1ASSOCIATION.SRC),b) for x in self.mdb.Read(a,STR(M1OBJECT.DSTASSOCIATIONS),b)])
        
        root = NON()
        if(0<len(args)):
            root = self.EvalOnContextAndScope(context,args[0],context,history)
        #TODO: Associates does not understand scope
        res = ApplyToAllElementsInA(context,root,asoFunctor)
        return res
    
    def EvalMetIdx(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CFT,M1OBJECT.CHILDCOMPOSITION,M1COMPOSITION.POS))
        def idxFunctor(a,b): 
            cc = self.mdb.Read(a,STR(M1OBJECT.CHILDCOMPOSITION))
            return self.mdb.Read(cc,STR(M1COMPOSITION.POS))
        res = ApplyToAllElementsInA(context,NON(),idxFunctor)
        return res
    
    def EvalMetCft(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CFT,M1OBJECT.CHILDCOMPOSITION,M1COMPOSITION.M2COMPOSITION))
        def cftFunctor(a,b): 
            cc = self.mdb.Read(a,STR(M1OBJECT.CHILDCOMPOSITION))
            return self.mdb.Read(cc,STR(M1COMPOSITION.M2COMPOSITION))
        res = ApplyToAllElementsInA(context,NON(),cftFunctor)
        return res
    
    def EvalMetFea(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/[%s,%s,%s] instead.'%(MET_MODES.FEA,M1OBJECT.M2CLASS,M2CLASS.MYATTRIBUTES,M2CLASS.MYSRCASSOCIATIONS,M2CLASS.MYPARENTCOMPOSITIONS))
        feaFunctor = lambda a,b:    self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.MYATTRIBUTES))+\
                                    self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.MYSRCASSOCIATIONS))+\
                                    self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.MYPARENTCOMPOSITIONS))
        
        res = ApplyToAllElementsInA(context,NON(),feaFunctor)
        return res
    
    def EvalMetFen(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/[%s,%s,%s] instead.'%(MET_MODES.FEN,M1OBJECT.M2CLASS,M2CLASS.ATTRIBUTES,M2CLASS.SRCASSOCIATIONS,M2CLASS.PARENTCOMPOSITIONS))
        fenFunctor = lambda a,b:    LST([self.mdb.Read(x,STR(M2ATTRIBUTE.NAME)) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.ATTRIBUTES))])+\
                                    LST([self.mdb.Read(x,STR(M2ASSOCIATION.DSTNAME)) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.SRCASSOCIATIONS))])+\
                                    LST([self.mdb.Read(x,STR(M2COMPOSITION.NAME)) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.PARENTCOMPOSITIONS))])
                                    
        res = ApplyToAllElementsInA(context,None,fenFunctor)
        return res
    
    def EvalMetFev(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/[%s,%s,%s] instead.'%(MET_MODES.FEV,M1OBJECT.M2CLASS,M2CLASS.MYATTRIBUTES,M2CLASS.MYSRCASSOCIATIONS,M2CLASS.PARENTCOMPOSITIONS))
        fevFunctor = lambda a,b:    LST([self.mdb.Read(a,self.mdb.Read(x,STR(M2ATTRIBUTE.NAME))) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.ATTRIBUTES))])+\
                                    LST([self.mdb.Read(a,self.mdb.Read(x,STR(M2ASSOCIATION.DSTNAME))) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.SRCASSOCIATIONS))])+\
                                    LST([self.mdb.Read(a,self.mdb.Read(x,STR(M2COMPOSITION.NAME))) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.PARENTCOMPOSITIONS))])
                                    
        res = ApplyToAllElementsInA(context,NON(),fevFunctor)
        return res
    
    def EvalMetAtt(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.ATT,M1OBJECT.M2CLASS,M2CLASS.ATTRIBUTES))
        attFunctor = lambda a,b: self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.ATTRIBUTES))
        res = ApplyToAllElementsInA(context,None,attFunctor)
        return res
    
    def EvalMetAtn(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s/%s instead.'%(MET_MODES.ATN,M1OBJECT.M2CLASS,M2CLASS.ATTRIBUTES,M2ATTRIBUTE.NAME))
        atnFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2ATTRIBUTE.NAME)) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.ATTRIBUTES))])
        res = ApplyToAllElementsInA(context,NON(),atnFunctor)
        return res
    
    def EvalMetAtv(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.ATV,M1OBJECT.M2CLASS,M2CLASS.ATTRIBUTES))
        atvFunctor = lambda a,b: LST([self.mdb.Read(a,self.mdb.Read(x,STR(M2ATTRIBUTE.NAME))) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.ATTRIBUTES))])
        res = ApplyToAllElementsInA(context,NON(),atvFunctor)
        return res
    
    def EvalMetRef(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.REF,M1OBJECT.M2CLASS,M2CLASS.SRCASSOCIATIONS))
        refFunctor = lambda a,b: self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.SRCASSOCIATIONS))
        res = ApplyToAllElementsInA(context,NON(),refFunctor)
        return res
    
    def EvalMetRen(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s/%s instead.'%(MET_MODES.ATV,M1OBJECT.M2CLASS,M2CLASS.SRCASSOCIATIONS,M2ASSOCIATION.DSTNAME))
        renFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2ASSOCIATION.DSTNAME)) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.SRCASSOCIATIONS))])
        res = ApplyToAllElementsInA(context,NON(),renFunctor)
        return res
    
    def EvalMetRev(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.REV,M1OBJECT.M2CLASS,M2CLASS.SRCASSOCIATIONS))
        revFunctor = lambda a,b: LST([self.mdb.Read(a,self.mdb.Read(x,STR(M2ASSOCIATION.DSTNAME))) for x in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.SRCASSOCIATIONS))])
        res = ApplyToAllElementsInA(context,NON(),revFunctor)
        return res
    
    def EvalMetCnt(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.CNT,M1OBJECT.M2CLASS,M2CLASS.PARENTCOMPOSITIONS))
        cntFunctor = lambda a,b: self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.PARENTCOMPOSITIONS))
        res = ApplyToAllElementsInA(context,NON(),cntFunctor)
        return res
    
    def EvalMetCnn(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CNN,M2CLASS.MYPARENTCOMPOSITIONS,M2COMPOSITION.NAME))
        cnnFunctor = lambda a,b : LST([self.mdb.Read(c,STR(M2COMPOSITION.NAME)) for c in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.MYPARENTCOMPOSITIONS))])
        res = ApplyToAllElementsInA(context,NON(),cnnFunctor)
        return res
    
    def EvalMetCnv(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s/%s instead.'%(MET_MODES.CNV,M1OBJECT.M2CLASS,M2CLASS.MYPARENTCOMPOSITIONS))
        cnvFunctor = lambda a,b: LST([self.mdb.Read(a,self.mdb.Read(c,STR(M2COMPOSITION.NAME))) for c in self.mdb.Read(self.mdb.Read(a,STR(M1OBJECT.M2CLASS)),STR(M2CLASS.MYPARENTCOMPOSITIONS))])
        res = ApplyToAllElementsInA(context,NON(),cnvFunctor)
        return res
    
    def EvalMetPac(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use /%s instead.'%(MET_MODES.CNT,M2CLASS.PACKAGE))
        pacFunctor = lambda a,b: self.mdb.Read(a,STR(M2CLASS.PACKAGE))
        res = ApplyToAllElementsInA(context,NON(),pacFunctor)
        return res
    
    def EvalMetSty(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CNT,M2CLASS.GENERALIZATIONS,M2INHERITANCE.SUPERCLASS))
        styFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2INHERITANCE.SUPERCLASS)) for x in self.mdb.Read(a,STR(M2CLASS.GENERALIZATIONS))])
        res = ApplyToAllElementsInA(context,NON(),styFunctor)
        return res
    
    def EvalMetAls(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CNT,M2CLASS.GENERALIZATIONS,M2INHERITANCE.SUPERCLASS))
        alsFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2INHERITANCE.SUPERCLASS)) for x in self.mdb.Read(a,STR(M2CLASS.GENERALIZATIONS))])
        res = ApplyToAllElementsInA(context,NON(),alsFunctor)
        return res
    
    def EvalMetImp(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CNT,M2CLASS.SPECIALIZATIONS,M2INHERITANCE.SUBCLASS))
        impFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2INHERITANCE.SUBCLASS)) for x in self.mdb.Read(a,STR(M2CLASS.SPECIALIZATIONS))])
        res = ApplyToAllElementsInA(context,NON(),impFunctor)
        return res
    
    def EvalMetAli(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s/%s instead.'%(MET_MODES.CNT,M2CLASS.SPECIALIZATIONS,M2INHERITANCE.SUBCLASS))
        aliFunctor = lambda a,b: LST([self.mdb.Read(x,STR(M2INHERITANCE.SUBCLASS)) for x in self.mdb.Read(a,STR(M2CLASS.SPECIALIZATIONS))])
        res = ApplyToAllElementsInA(context,NON(),aliFunctor)
        return res
    
    def EvalMetMmo(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        self.logger.Warn('DEPRECATED: @%s is deprecated, please use %s.%s instead.'%(MET_MODES.MMO,CONCEPTS.MXMDB,MXMDB.M2MODELS))
        (mdb,perr) = self.mdb.Create(STR(CONCEPTS.MXMDB))
        res = self.mdb.Read(mdb,STR(MXMDB.M2MODELS)) 
        return res
    
    
    def EvalMetIff(self, context:VAL, args:LST, scope:VAL, history:History) -> VAL:
        
        condition = self.EvalOnContextAndScope(context,args[0],context,history)
        
        res = NON()
        if(condition.IsTrue()):
            res = self.EvalOnContextAndScope(context,args[1],context,history)
        else:
            res = self.EvalOnContextAndScope(context,args[2],context,history)
        return res
    
    '''
        LOGICAL AND MATH OPERATORS
    '''
    
    def EvalEqu(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.EQU, self.equEvaluators,history)
        return res
    
    def EvalEqa(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        #local functor
        def eqaFunctor(a,b):
            return BOL(a in b)
        #method start
        select = self.EvalOnContextAndScope(context,args[0],context,history)
        if(not IsListOfObjects(select)):
            raise EOQ_ERROR_INVALID_TYPE("Argument must be a list of elements but got: %s"%(select))
        res = ApplyToAllElementsInA(context,select,eqaFunctor)
        return res
    
    def EvalNeq(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.NEQ, self.neqEvaluators,history)
        return res
    
    def EvalLes(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.LES, self.lesEvaluators,history)
        return res
    
    def EvalGre(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.GRE, self.greEvaluators,history)
        return res
    
    def EvalRgx(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        regex = ValidateVal(self.EvalOnContextAndScope(context,args[0],context,history),[STR],'regex')
        if(not isinstance(regex,STR)):
            raise EOQ_ERROR_INVALID_TYPE("Argument must be STR, but got %s."%(args))
        pattern = None
        try:
            pattern = RGX(re.compile(regex.GetVal()))
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("%s is no valid regular expression: %s"%(regex,str(e)))
        res = self.__EvalElementOperation(context, pattern, scope, SEG_TYPES.RGX, self.rgxEvaluators,history)
        return res
    
    def EvalAdd(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.ADD, self.addEvaluators,history)
        return res
    
    def EvalSub(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.SUB, self.subEvaluators,history)
        return res
    
    def EvalMul(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.MUL, self.mulEvaluators,history)
        return res
    
    def EvalDiv(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalElementOperation(context, args[0], scope, SEG_TYPES.DIV, self.divEvaluators,history)
        return res
    
    
    def EvalCsp(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalListOfElementsOperation(context, args[0], scope, SEG_TYPES.CSP, self.cspEvaluator,history)
        return res
    
    def EvalIts(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalListOfElementsOperation(context, args[0], scope, SEG_TYPES.ITS, self.itsEvaluator,history)
        return res
    
    def EvalDif(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalListOfElementsOperation(context, args[0], scope, SEG_TYPES.ITS, self.difEvaluator,history)
        return res
    
    def EvalUni(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalListOfElementsOperation(context, args[0], scope, SEG_TYPES.UNI, self.uniEvaluator,history)
        return res
    
    def EvalCon(self, context:VAL, args:list, scope:VAL, history:History) -> VAL:
        res = self.__EvalListOfElementsOperation(context, args[0], scope, SEG_TYPES.CON, self.conEvaluator,history)
        return res
    
    '''
        PRIVATE METHODS
    '''
   
    def __EvalElementOperation(self, context:VAL, args:list, scope:VAL, operator, opEvaluators, history:History) -> VAL:  
        res = None
        #Define operators
        def opEqualListsFunc(a,b,c):
            return self.__ApplyOperator(a,b,opEvaluators)
        def opOnlyOp1ListFunc(a,b,c):
            op1Functor = lambda o1,o2: self.__ApplyOperator(o1,o2,opEvaluators)
            return ApplyToAllElementsInB(a,b,op1Functor)
        def opOnlyOp2ListFunc(a,b,c):
            op2Functor = lambda o1,o2: self.__ApplyOperator(o1,o2,opEvaluators)
            return ApplyToAllElementsInA(a,b,op2Functor)
    
        op1 = context
        op2 = self.EvalOnContextAndScope(context,args,scope,history)
        
        try:
            res = ApplyToSimilarElementStrutures(op1, op2, opEqualListsFunc, opOnlyOp1ListFunc, opOnlyOp2ListFunc)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Failed to evaluate %s. Context and arguments must be single elements or arrays of same type and size, but got %s %s %s: %s"%(operator,op1,operator,op2,str(e)))
        return res

    def __EvalListOfElementsOperation(self, context:VAL, args:list, scope:VAL, operator, opEvaluator, history:History) -> VAL:  
        res = None
        #Define operators
        def opEqualListsFunc(a,b,c):
            return opEvaluator(a,b)
        def opOnlyOp1ListFunc(a,b,c):
            return ApplyToAllElementsInB(a,b,opEvaluator)
        def opOnlyOp2ListFunc(a,b,c):
            return ApplyToAllElementsInA(a,b,opEvaluator)
    
        op1 = context
        op2 = self.EvalOnContextAndScope(context,args,scope,history)
        
        try:
            res = ApplyToSimilarListsOfObjects(op1, op2, opEqualListsFunc, opOnlyOp1ListFunc, opOnlyOp2ListFunc)
        except Exception as e:
            raise EOQ_ERROR_INVALID_VALUE("Failed to evaluate %s. Context and arguments must be single elements or arrays of same type and size, but got %s %s %s: %s"%(operator,op1,operator,op2,str(e)))
        return res
                
    def __ApplyOperator(self,a,b,operatorHandlers):
        aType = a.Type()
        bType = b.Type()
        try:
            handler = operatorHandlers[(aType,bType)]
            if(None == handler):
                raise EOQ_ERROR_INVALID_TYPE('Unsupported operation for types %s and %s.'%(aType,bType))
            return handler(a,b)
        except KeyError:
            raise EOQ_ERROR_INVALID_TYPE('Unsupported operation for types %s and %s.'%(aType,bType))
        
    def __Flatten(self,src : VAL,target : LST) -> LST:
        for x in src:
            if(IsList(x)):
                self.__Flatten(x, target)
            else: 
                target.append(x)
        
    def __InitQrySegHandlers(self):
        handlerTable = {}
        handlerTable[SEG_TYPES.OBJ] = self.EvalObj
        handlerTable[SEG_TYPES.ONI] = self.EvalOni
        handlerTable[SEG_TYPES.HIS] = self.EvalHis
        
        handlerTable[SEG_TYPES.PTH] = self.EvalPth
        handlerTable[SEG_TYPES.CLS] = self.EvalCls
        handlerTable[SEG_TYPES.INO] = self.EvalIno
        handlerTable[SEG_TYPES.MET] = self.EvalMet
        handlerTable[SEG_TYPES.NOT] = self.EvalNot
        handlerTable[SEG_TYPES.TRM] = self.EvalTrm
        handlerTable[SEG_TYPES.TRY] = self.EvalTry
        
        handlerTable[SEG_TYPES.QRF] = self.EvalQrf
        handlerTable[SEG_TYPES.STF] = self.EvalStf
        handlerTable[SEG_TYPES.SLF] = self.EvalSlf
        
        handlerTable[SEG_TYPES.IDX] = self.EvalIdx
        handlerTable[SEG_TYPES.SEL] = self.EvalSel
        handlerTable[SEG_TYPES.ARR] = self.EvalArr
        handlerTable[SEG_TYPES.ZIP] = self.EvalZip
        
        handlerTable[SEG_TYPES.ANY] = self.EvalAny
        handlerTable[SEG_TYPES.ALL] = self.EvalAll
        
        handlerTable[SEG_TYPES.EQU] = self.EvalEqu
        handlerTable[SEG_TYPES.EQA] = self.EvalEqa
        handlerTable[SEG_TYPES.NEQ] = self.EvalNeq
        handlerTable[SEG_TYPES.LES] = self.EvalLes
        handlerTable[SEG_TYPES.GRE] = self.EvalGre
        handlerTable[SEG_TYPES.RGX] = self.EvalRgx
        
        handlerTable[SEG_TYPES.ADD] = self.EvalAdd
        handlerTable[SEG_TYPES.SUB] = self.EvalSub
        handlerTable[SEG_TYPES.MUL] = self.EvalMul
        handlerTable[SEG_TYPES.DIV] = self.EvalDiv
        
        #synonyms for boolean operations
        handlerTable[SEG_TYPES.ORR] = self.EvalAdd
        handlerTable[SEG_TYPES.XOR] = self.EvalSub
        handlerTable[SEG_TYPES.AND] = self.EvalMul
        handlerTable[SEG_TYPES.NAD] = self.EvalDiv
        
        handlerTable[SEG_TYPES.CSP] = self.EvalCsp
        handlerTable[SEG_TYPES.ITS] = self.EvalIts
        handlerTable[SEG_TYPES.DIF] = self.EvalDif
        handlerTable[SEG_TYPES.UNI] = self.EvalUni
        handlerTable[SEG_TYPES.CON] = self.EvalCon
        return handlerTable
        
    def __InitMetOperationHandlers(self):
        handlerTable = {}
        handlerTable[MET_MODES.CLS] = self.EvalMetCls
        handlerTable[MET_MODES.CLN] = self.EvalMetCln
        handlerTable[MET_MODES.LEN] = self.EvalMetLen
        handlerTable[MET_MODES.PAR] = self.EvalMetPar
        handlerTable[MET_MODES.CON] = self.EvalMetPar #container is the same as parent
        handlerTable[MET_MODES.ALP] = self.EvalMetAlp
        handlerTable[MET_MODES.ASO] = self.EvalMetAso
        handlerTable[MET_MODES.IDX] = self.EvalMetIdx
        handlerTable[MET_MODES.CFT] = self.EvalMetCft
        handlerTable[MET_MODES.FEA] = self.EvalMetFea
        handlerTable[MET_MODES.FEN] = self.EvalMetFen
        handlerTable[MET_MODES.FEV] = self.EvalMetFev
        handlerTable[MET_MODES.ATT] = self.EvalMetAtt
        handlerTable[MET_MODES.ATN] = self.EvalMetAtn
        handlerTable[MET_MODES.ATV] = self.EvalMetAtv
        handlerTable[MET_MODES.REF] = self.EvalMetRef
        handlerTable[MET_MODES.REN] = self.EvalMetRen
        handlerTable[MET_MODES.REV] = self.EvalMetRev
        handlerTable[MET_MODES.CNT] = self.EvalMetCnt
        handlerTable[MET_MODES.CNN] = self.EvalMetCnn
        handlerTable[MET_MODES.CNV] = self.EvalMetCnv
        #class meta operators
        handlerTable[MET_MODES.PAC] = self.EvalMetPac
        handlerTable[MET_MODES.STY] = self.EvalMetSty
        handlerTable[MET_MODES.ALS] = self.EvalMetAls
        handlerTable[MET_MODES.IMP] = self.EvalMetImp
        handlerTable[MET_MODES.ALI] = self.EvalMetAli
        handlerTable[MET_MODES.MMO] = self.EvalMetMmo
        #control flow operators
        handlerTable[MET_MODES.IFF] = self.EvalMetIff
        return handlerTable
    
    
    def __InitOpHandlerEnt(self,aType,bolHandler,\
                           u32Handler,u64Handler,\
                           i32Handler,i64Handler,\
                           f32Handler,f64Handler,\
                           strHandler,datHandler,\
                           qryHandler,\
                           nonHandler,trmHandler):
        return {(aType,BOL):bolHandler,
                (aType,U32):u32Handler,
                (aType,U64):u64Handler,
                (aType,I32):i32Handler,
                (aType,I64):i64Handler,
                (aType,F32):f32Handler,
                (aType,F64):f64Handler,
                (aType,STR):strHandler,
                (aType,DAT):datHandler,
                (aType,QRY):qryHandler,
                (aType,NON):nonHandler,
                (aType,TRM):trmHandler}
            
        
    def __InitEquOperationHandlers(self):
        handlerTable = {}
        #handler functors
        XvX = lambda a,b: BOL(a==b) #any combination
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, XvX , None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    
    def __InitNeqOperationHandlers(self):
        handlerTable = {}
        #handler functors
        XvX = lambda a,b: BOL(a!=b) #any combination
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    
    def __InitGreOperationHandlers(self):
        handlerTable = {}
        #handler functors
        XvX = lambda a,b: BOL(a>b) #any combination
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, XvX , None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    
    def __InitLesOperationHandlers(self):
        handlerTable = {}
        #handler functors
        XvX = lambda a,b: BOL(a<b) #any combination
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, XvX , None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
        
    def __InitRgxOperationHandlers(self):
        handlerTable = {}
        #handler functors
        XvS = lambda a,b: BOL(True) if(b.GetVal().search(a.GetVal())) else BOL(False) #only string can be the second argument. 
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvS , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    
    def __InitAddOperationHandlers(self):
        handlerTable = {}
        #handler functors STR > BOL > DBL > FLO > INT > OBJ > NON
        XvX = lambda a,b: a+b #only for numeric types, i.e. INT, FLO, DBL
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    
    def __InitSubOperationHandlers(self):
        handlerTable = {}
        #handler functors STR > BOL > DBL > FLO > INT > OBJ > NON
        XvX = lambda a,b: a-b #only for numeric types, i.e. INT, FLO, DBL
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, XvX , None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable

       
    def __InitMulOperationHandlers(self):
        handlerTable = {}
        #handler functors STR > BOL > DBL > FLO > INT > OBJ > NON
        XvX = lambda a,b: a*b #only for numeric types, i.e. INT, FLO, DBL
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
        
    def __InitDivOperationHandlers(self):
        handlerTable = {}
        #handler functors STR > BOL > DBL > FLO > INT > OBJ > NON
        XvX = lambda a,b: a/b #only for numeric types, i.e. INT, FLO, DBL
        # create a handler entry and compatibility list for every primitive type combination.
        # None indicates incompatible operations      
        #                                           first operator type        , BOL , INT , FLO , DBL , STR , OBJ , NON , TRM
                # first operator type                            BOL , U32 , U64 , I32 , I64 , F32 , F64 , STR , DAT , QRY , NON , TRM
        handlerTable.update(self.__InitOpHandlerEnt(BOL, XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , XvX , None))
        handlerTable.update(self.__InitOpHandlerEnt(U32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(U64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(I64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F32, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(F64, None, XvX , XvX , XvX , XvX , XvX , XvX , None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(STR, None, None, None, None, None, None, None, XvX , None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(DAT, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(QRY, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(NON, None, None, None, None, None, None, None, None, None, None, None, None))
        handlerTable.update(self.__InitOpHandlerEnt(TRM, TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX , TvX )) 
        return handlerTable
    
    def __DeprecatedWarning(self, message:str):
        self.logger.Warn("%s is deprecated. Consider replacement."%(message))
                
        
                


'''
Definition of value types used in EOQ in order to make sure that EOQ is type safe in all target languages and serialization.
 
Bjoern Annighoefer 2022
'''

from ..error import EOQ_ERROR_INVALID_TYPE, EOQ_ERROR_INVALID_VALUE, EOQ_ERROR_UNSUPPORTED, EOQ_ERROR_RUNTIME
from datetime import datetime,timezone,timedelta
from math import floor
#type checking
from typing import Any, Dict, Union, Type, List, Tuple, Iterable
from abc import ABC


### START OF EOQ VALUE TYPES DEFINITION ###

class VAL_TYPES:
    BOL = 'BOL' #boolean
#     U8  = 'U8'  #byte
#     U16 = 'U16' #word
    U32 = 'U32' #unsigned nit 32 bit #NOT IMPLEMENTED
    U64 = 'U64' #unsigned nit 64 bit #NOT IMPLEMENTED
#     I16 = 'I16' #signed int 16 bit (short)
    I32 = 'I32' #signed int 32 bit (long)
    I64 = 'I64' #signed int 64 bit (long long)
    F32 = 'F32' #floating point 32 bit
    F64 = 'F64' #floating point 64 bit (double)
    STR = 'STR' #String
    DAT = 'DAT' #date and time #NOT IMPLEMENTED
    QRY = 'QRY' #query
    NON = 'NON' # none or null
    LST = 'LST' #list

### BASE CLASS FOR ALL VALUES ###
    
class VAL(ABC):
    '''The base class for any value in EOQ 
    
    '''
    def __init__(self, valType:str):
        self.val = valType #the json serializer relies on this
    # PYTHON SPECIFIC
    def __bool__(self):
        return self.IsTrue()
    def __eq__(self, other):
        if(self.Type() == type(other)):
            return (self.v == other.v)
        elif(None == other): #handles general None compares
            return self.IsNone()
        else:
            try:
                #try casting the value types to compare them.
                return (self.v == self.Cast(other).v)
            except:
                return False #everything that cannot be casted is not the same
    def __lt__(self, other):
        if(isinstance(other,VAL) and self.Type() == other.Type()):
            return (self.v < other.v)
        else:
            return self.v < self.Cast(other).v
    def __gt__(self, other):
        if(isinstance(other,VAL) and self.Type() == other.Type()):
            return (self.v > other.v)
        else:
            return self.v > self.Cast(other).v
    def __le__(self, other):
        return not (self > other)
    def __ge__(self, other):
        return not (self < other)
    def __hash__(self):
        return hash(self.__repr__())
    # VAL INTERFACE
    def Type(self):
        '''Returns the value type. 
        This is an own function to make it overridable, which makes sence in case of QRY
        '''
        return type(self) ##by default the class is returned
    def Cast(self, other):
        ''' Casts from any possible value to this value type
        Raises an exception if the value type cannot be casted
        '''
        T = self.Type()
        return other if isinstance(other,T) else T(other)
    def IsNone(self):
        '''Shall be used instead of none checks. 
        By default values are never None, except NON.
        
        '''
        return False 
    def IsTrue(self):
        '''Shall be used in if conditions like instead of only using the value
        
        '''
        return True if(self.v) else False
    def GetVal(self):
        '''Return the native value of the native programming language
        E.g I32 might return int in Python and Numeric in Javascript 
        '''
        return self.v #default implementation, but maybe overridden
    def SetVal(self,val) -> None:
        '''Set from the native value
        
        '''
        self.v = val #default implementation might be overridden
        
    


### VALUE REGISTRY ###

class ValInfo:
    def __init__(self, valType:str, description:str, clazz:Type[VAL], unpackArgs:bool):
        self.valType:str     = valType
        self.description:str = description
        self.clazz:Type[VAL] = clazz
        self.unpackArgs:bool = unpackArgs
        

VAL_TYPE_REGISTRY:Dict[str,ValInfo] = {}
    

def RegisterValType(valType:str, description:str, clazz:Type[VAL], unpackArgs:bool=False)->None:
    if(valType in VAL_TYPE_REGISTRY):
        raise EOQ_ERROR_RUNTIME('Value type %s is already registered.'%(valType))
    VAL_TYPE_REGISTRY[valType] = ValInfo(valType, description, clazz, unpackArgs)


### BASE CLASSES FOR PRIMITIVE VALUES ###
class SIV(VAL):
    '''A single value data type
    
    '''


class PRM(SIV):
    '''A primitive type. This covers all except list
    
    '''
    
class NUM(PRM):
    '''All numeric values
    
    '''
    # PYTHON SPECIFIC
    def __add__(self,b):
        return self.Type()(self.v + self.Cast(b).v)
    def __sub__(self,b):
        return self.Type()(self.v - self.Cast(b).v)
    def __mul__(self,b):
        return self.Type()(self.v * self.Cast(b).v)
    def __truediv__(self,b):
        try:
            return self.Type()(self.v / self.Cast(b).v)
        except ZeroDivisionError:
            raise EOQ_ERROR_INVALID_VALUE('Division by zero')
    def __neg__(self):
        return self.Type()(-self.v)
    
class INT(NUM):
    '''All integer values
    
    '''
    
class FLO(NUM):
    '''All floating point values
    
    '''

      
### PRIMITIVE TYPES #####
        
class BOL(PRM):
    def __init__(self, value:Union[bool,VAL]=False):
        VAL.__init__(self,VAL_TYPES.BOL)
        self.v:bool = False
        if(bool == type(value)):
            self.v = value
        elif(isinstance(value,VAL)):
            self.v = value.IsTrue()
        else:
            self.v = True if value else False
    # PYTHON SPECIFIC
    def __repr__(self):
        return "b1" if self.v else "b0"
    # PYTHON SPECIFIC
    def __add__(self,b): #OR
        a = self.v; b = self.Cast(b).v
        return self.Type()(a or b)
    def __sub__(self,b): #XOR
        a = self.v; b = self.Cast(b).v
        return self.Type()((not a and b) or (a and not b))
    def __mul__(self,b): #AND
        a = self.v; b = self.Cast(b).v
        return self.Type()(a and b)
    def __truediv__(self,b): #NAND
        a = self.v; b = self.Cast(b).v
        return self.Type()(not (a and b))
    def __neg__(self):
        return self.Type()(not self.v)
RegisterValType(VAL_TYPES.BOL,'Boolean',BOL)

class U32(INT):
    def __init__(self, value:Union[int,float,INT,FLO]=0):
        VAL.__init__(self,VAL_TYPES.U32)
        self.v:int = 0
        if(int == type(value)):
            self.SetVal(value)
        elif(float == type(value)):
            self.SetVal(round(value))
        elif(isinstance(value,U32) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,INT) and not value.IsNone()):
            self.SetVal(value.GetVal())
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.SetVal(round(value.GetVal()))
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))
    # PYTHON SPECIFIC
    def __repr__(self):
        return "u%d"%(self.v)
    # VAL INTERFACE
    def SetVal(self,val:int) -> None:
        '''simulate u32 value space '''
        minVal = 0
        maxVal = 4294967296-1#2^32-1
        if(val < minVal):
            raise EOQ_ERROR_INVALID_VALUE("Underflow: value can not be smaller than %d, but got %d."%(minVal,val))
        elif(maxVal < val):
            raise EOQ_ERROR_INVALID_VALUE("Overflow: value can not be larger than %d, but got %d."%(maxVal,val))
        self.v = val
#         maxVal = 4294967296#2^32
#         #check if value is negative
#         if(0>val):
#             self.v = maxVal - ((-val) % maxVal) #simulate 2-complement
#         else:
#             self.v = val % maxVal #wrap around
RegisterValType(VAL_TYPES.U32,'32bit unsigned integer',U32)

class U64(INT):
    def __init__(self, value:Union[int,float,INT,FLO]=0):
        VAL.__init__(self,VAL_TYPES.U64)
        self.v:int = 0
        if(int == type(value)):
            self.SetVal(value)
        elif(float == type(value)):
            self.SetVal(round(value))
        elif(isinstance(value,U64) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,INT) and not value.IsNone()):
            self.SetVal(value.GetVal())
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.SetVal(round(value.GetVal()))
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))
    # PYTHON SPECIFIC    
    def __repr__(self):
        return "u%d"%(self.v)
    # VAL INTERFACE
    def SetVal(self,val:int) -> None:
        '''simulate u64 value space '''
        minVal = 0
        maxVal = 18446744073709551616-1#2^32-1
        if(val < minVal):
            raise EOQ_ERROR_INVALID_VALUE("Underflow: value can not be smaller than %d, but got %d."%(minVal,val))
        elif(maxVal < val):
            raise EOQ_ERROR_INVALID_VALUE("Overflow: value can not be larger than %d, but got %d."%(maxVal,val))
        self.v = val
RegisterValType(VAL_TYPES.U64,'64bit unsigned integer',U64)
    
class I32(INT):
    def __init__(self, value:Union[int,float,INT,FLO]=0):
        VAL.__init__(self,VAL_TYPES.I32)
        self.v:int = 0
        if(int == type(value)):
            self.SetVal(value)
        elif(float == type(value)):
            self.SetVal(round(value))
        elif(isinstance(value,I32) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,INT) and not value.IsNone()):
            self.SetVal(value.GetVal())
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.SetVal(round(value.GetVal()))
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))
    # PYTHON SPECIFIC     
    def __repr__(self):
        return "i%d"%(self.v)
    # VAL INTERFACE
    def SetVal(self, val:int) -> None:
        '''simulate i32 value space '''
        minVal = -2147483648 #-2^31
        maxVal = 2147483648-1#2^31-1
        if(val < minVal):
            raise EOQ_ERROR_INVALID_VALUE("Underflow: value can not be smaller than %d, but got %d."%(minVal,val))
        elif(maxVal < val):
            raise EOQ_ERROR_INVALID_VALUE("Overflow: value can not be larger than %d, but got %d."%(maxVal,val))
        self.v = val
RegisterValType(VAL_TYPES.I32,'32bit signed integer',I32)


class I64(INT):
    def __init__(self, value:Union[int,float,INT,FLO]=0):
        VAL.__init__(self,VAL_TYPES.I64)
        self.v:int = 0
        if(int == type(value)):
            self.SetVal(value)
        elif(float == type(value)):
            self.SetVal(round(value))
        elif(isinstance(value,I64) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,INT) and not value.IsNone()):
            self.SetVal(value.GetVal())
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.SetVal(round(value.GetVal()))
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))  
    def __repr__(self):
        return "l%d"%(self.v)
    # VAL INTERFACE
    def SetVal(self, val:int) -> None:
        '''simulate i64 value space '''
        minVal = -9223372036854775808 #-2^31
        maxVal = 9223372036854775808-1#2^31-1
        if(val < minVal):
            raise EOQ_ERROR_INVALID_VALUE("Underflow: value can not be smaller than %d, but got %d."%(minVal,val))
        elif(maxVal < val):
            raise EOQ_ERROR_INVALID_VALUE("Overflow: value can not be larger than %d, but got %d."%(maxVal,val))
        self.v = val
RegisterValType(VAL_TYPES.I64,'64bit signed integer',I64)
    
class F32(FLO):
    def __init__(self, value:Union[float,int,FLO,INT]=0.0):
        VAL.__init__(self,VAL_TYPES.F32)
        self.v:float = 0.0
        if(float == type(value)):
            self.v = value
        elif(int == type(value)):
            self.v = float(value)
        elif(isinstance(value,F32) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.v = value.GetVal()
        elif(isinstance(value,INT) and not value.IsNone()):
            self.v = float(value.GetVal())
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))
    # PYTHON SPECIFIC         
    def __repr__(self):
        return "f%f"%(self.v)
RegisterValType(VAL_TYPES.F32,'32bit floating point (float)',F32)

class F64(FLO):
    def __init__(self, value:Union[float,int,FLO,INT]=0.0):
        VAL.__init__(self,VAL_TYPES.F64)
        self.v:float = 0.0
        if(float == type(value)):
            self.v = value
        elif(int == type(value)):
            self.v = float(value)
        elif(isinstance(value,F64) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        elif(isinstance(value,FLO) and not value.IsNone()):
            self.v = value.GetVal()
        elif(isinstance(value,INT) and not value.IsNone()):
            self.v = float(value.GetVal())
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected int, float, INT or FLO but got %s."%(value))
    # PYTHON SPECIFIC         
    def __repr__(self):
        return "d%f"%(self.v)
RegisterValType(VAL_TYPES.F64,'64bit floating point (double)',F64)
    
class STR(PRM):
    def __init__(self, value:Union[str,VAL]=''):
        VAL.__init__(self,VAL_TYPES.STR)
        self.v:str = ''
        if(str == type(value)):
            self.v = value
        elif(isinstance(value,STR) and not value.IsNone()): ##Copy constructor
            self.v = value.v
        else:
            raise EOQ_ERROR_INVALID_TYPE("Expected str or STR, but got %s."%(value))
    # PYTHON SPECIFIC         
    def __repr__(self):
        return self.v  
    def __len__(self):
        return len(self.v)
    def __getitem__(self,key:Union[int,slice,INT]):
        if(int == type(key)):
            return STR(self.v[key])
        elif(isinstance(key,slice)):
            return STR(self.v[key])
        elif(isinstance(key,INT) and not key.IsNone()):
            return STR(self.v[key.GetVal()])
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected int, slice or INT but got %s.'%(key))
    def __add__(self, other):
        return STR(self.v + self.Cast(other).v)
RegisterValType(VAL_TYPES.STR,'String',STR)

class DAT(PRM):
    def __init__(self, yearOrValue:Union[int,datetime,VAL]=0, month:int=0, day:int=0, hour:int=0, minute:int=0, second:int=0, microsec:float=0.0, utcoff:float=0.0):
        VAL.__init__(self,VAL_TYPES.DAT)
        self.v:Tuple[int,int,int,int,int,int,float,float] = (0,0,0,0,0,0,0.0,0.0)
        if(isinstance(yearOrValue,datetime)):
            self.SetVal(yearOrValue)
        elif(DAT == type(yearOrValue)):
            self.v = tuple(yearOrValue.v)
        elif(int == type(yearOrValue)):
            self.v = (yearOrValue,month,day,hour,minute,second,microsec,utcoff)
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected datetime, DAT or year as int but got %s.'%(yearOrValue))
    # PYTHON SPECIFIC
    def __repr__(self):
        # ISO 8601 format: YYYY-MM-DDThh:mm:ss.sss+hh:mm (https://en.wikipedia.org/wiki/ISO_8601)
        Y = self.v[0]
        M = self.v[1]
        D = self.v[2]
        h = self.v[3]
        m = self.v[4]
        s = self.v[5]
        ms = int(round(self.v[6]*1000)) # convert from us to full ms
        oh = int(floor(self.v[7])) #extract full h
        om = int(round((self.v[7] - oh)*60)) #extract full min
        return "%04d-%02d-%02dT%02d:%02d:%02d:%f%0+2d:%02d"%(Y,M,D,h,m,s,ms,oh,om)
    def __lt__(self, other):
        return self.GetVal() < self.Cast(other).GetVal()
    def __gt__(self, other):
        return self.GetVal() > self.Cast(other).GetVal()
    # VAL INTERFACE
    def GetVal(self) -> datetime:
        tz = timezone(timedelta(seconds=int(self.utcoff()*3600.0)))
        return datetime(self.year(),self.month(),self.day(),self.hour(),self.minute(),self.second(),int(self.microsec()),tz)
    def SetVal(self,val : datetime):
        utcoff = 0.0
        td = val.utcoffset() #timedelta
        if(td): #the timezone info is not always present
            h,r = divmod(td.seconds,3600)
            m,s = divmod(r,60)
            utcoff = float(h)+float(m)/60.0+float(s)/3600 #UTC offset in hours
        self.v = (val.year,val.month,val.day,val.hour,val.minute,val.second,val.microsecond,utcoff) #todo: timezone?
    # ADDITIONAL HELPER FUNCTIONS
    def year(self)->int:
        return self.v[0]
    def month(self)->int:
        return self.v[1]
    def day(self)->int:
        return self.v[2]
    def hour(self)->int:
        return self.v[3]
    def minute(self)->int:
        return self.v[4]
    def second(self)->int:
        return self.v[5]
    def microsec(self)->float:
        return self.v[6]
    def utcoff(self)->float:
        return self.v[7]
RegisterValType(VAL_TYPES.DAT,'Date and time',DAT,True)
    
class QRY(SIV):
    def __init__(self):
        VAL.__init__(self,VAL_TYPES.QRY)
        # is not implemented here but in query
    # VAL INTERFACE
    def GetVal(self):
        return self #query can not be resolved
    def SetVal(self,value)->None:
        raise EOQ_ERROR_UNSUPPORTED('QRY cannot be set')
    def Type(self):
        return QRY #make sure the inherited types also return QRY
    
# QRY is not registered because it is only a placeholder to initialize NON. The real query class is in query.py
    
# NON is something special, because every other value can also be none
class NON(BOL,I32,I64,F32,F64,STR,DAT,QRY):
    def __init__(self,value=None):
        VAL.__init__(self,VAL_TYPES.NON)
        self.v:list = []
        if(NON == type(value)):
            pass
        elif(None == value):
            pass
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected NON or None but got: %s.'%(value))
        #self.v = [] #empty list enables factory create better
    def __repr__(self):
        return "n0"
    # PYTHON SPECIFIC
    def __eq__(self, other):
        if(isinstance(other,VAL)):
            return other.IsNone()
        elif(None == other): #handles general None compares
            return True
        else:
            return False
    def __lt__(self, other):
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation < for NON')
    def __gt__(self, other):
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation > for NON')
    def __add__(self,b): #OR
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation + for NON')
    def __sub__(self,b): #XOR
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation - for NON')
    def __mul__(self,b): #AND
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation * for NON')
    def __truediv__(self,b): #NAND
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation / for NON')
    def __neg__(self):
        raise EOQ_ERROR_UNSUPPORTED('Unknown operation -x for NON')
    # VAL INTERFACE
    def IsNone(self):
        return True #this is the only None type
    def IsTrue(self):
        '''Override because NON values are always false'''
        return False
    # VAL INTERFACE
    def GetVal(self):
        '''Override because NON values are always None'''
        return None
    def SetVal(self,value)->None:
        pass #nothing. Cannot set NON to a different value
    def Type(self):
        return NON #make sure the inherited types also return QRY
RegisterValType(VAL_TYPES.NON,'None or null value',NON,True)
### COMPLEY TYPES ###

class LSTIterator:
    '''ListIterator is necessary to make LST iterable
    
    '''
    def __init__(self,lst):
        self.__lst = lst
        self.__i = 0 #start at zero
    def __next__(self):
        if(self.__i < len(self.__lst)):
            self.__i += 1
            return self.__lst[self.__i-1]
        else:
            raise StopIteration()
    def __iter__(self): #seems necessary for python >= 3.13
        return self
    
class LST(VAL):
    def __init__(self,value:Union[Iterable,VAL] = []):
        super().__init__(VAL_TYPES.LST)
        self.v:List[VAL] = []
        if(isinstance(value,Iterable)):
            #self.v = [v if isinstance(v,VAL) else raise EOQ_ERROR_INVALID_TYPE('Only VAL can be member of LST, but got %s'%(v)) for v in value ]
            self.v = []
            for v in value: 
                if(isinstance(v,VAL)):
                    self.v.append(v)
                else:
                    raise EOQ_ERROR_INVALID_TYPE('Only VAL can be member of LST, but got %s.'%(type(v).__name__))
        elif(isinstance(value,LST) and not value.IsNone()): #Copy constructor
            self.v = [x for x in value.v]
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected list or LST but got %s.'%(value))
    # PYTHON SPECIFIC    
    def __repr__(self):
        return str(self.v)
    def __len__(self):
        return len(self.v)
    def __getitem__(self,key:Union[int,slice,INT]):
        if(int == type(key)):
            return self.v[key]
        elif(isinstance(key,slice)):
            return LST(self.v[key])
        elif(isinstance(key,INT) and not key.IsNone()):
            return self.v[key.GetVal()]
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected int, slice or INT but got %s.'%(key))
    def __setitem__(self, key:int, newvalue:VAL):
        self.v[key] = newvalue
    def __iter__(self):
        return LSTIterator(self)
    def __contains__(self,item):
        return (item in self.v)
    def __add__(self, other:Union[list,'LST']):
        if(isinstance(other,LST)):
            return LST(self.v + other.v)
        else:
            return LST(self + self.Cast(other))
    ### simulate the list functions of python
    def append(self,val:VAL):
        if(isinstance(val,VAL)):
            self.v.append(val)
            return self
        else:
            raise EOQ_ERROR_INVALID_TYPE('Expected VAL but got %s.'%(val))
        
    def index(self,find):
        return self.v.index(find) 
    # VAL INTERFACE
    def GetVal(self)->list:
        return [v.GetVal() for v in self.v]
    def SetVal(self,value)->None:
        raise EOQ_ERROR_UNSUPPORTED('LST cannot be set')
RegisterValType(VAL_TYPES.LST,'list',LST)
    
#### VALUE INIT HELPERS ####

def EncVal(value:Union[VAL,None,bool,int,float,str,list]):
    '''Convert any primitive type in the corresponding EOQ value type
    
    '''
    if(None == value): #None checks are not handled via type compare
        return NON()
    elif(isinstance(value,VAL)):
        return value #nothing to do
    #all other types are handled by detecting the Python base type
    elif(bool == type(value)):
        return BOL(value)
    elif(int == type(value)):
        return I64(value)
    elif(float == type(value)):
        return F32(value)
    elif(str == type(value)):
        return STR(value)
    elif(list == type(value)):
        return LST([EncVal(v) for v in value])
    else:
        raise EOQ_ERROR_INVALID_TYPE("Data type %s is not supported."%(type(value).__name__))
    
def InitValOrNon(value,ValType):
    return  NON() if None==value else ValType(value)

### VALUE VALIDATION HELPERS ####
    
def ValidateVal(value:VAL, expectedTypes:list, varname:str, exact:bool=True)->VAL:
    '''Checks if the vale corresponds to a type from the list. 
    If not it raises a EOQ_ERROR_INVALID_TYPE exception with varname
    If exact is false it is checked for being an instance.
    '''
    for t in expectedTypes:
        if(exact and t == type(value) or isinstance(value,t)):
            return value
    #if going here all cast failed:
    expectedTypeStrs = " or ".join([t.__name__ for t in expectedTypes])
    raise EOQ_ERROR_INVALID_TYPE("%s: expected %s but got %s"%(varname,expectedTypeStrs,type(value).__name__)) 
    
def ValidateValAndCast(value:Union[VAL,bool,int,float,str,datetime,None], expectedTypes:List[Type[VAL]], varname:str)->VAL:
    '''Tries to cast the given value into the given VAL types after each other. The first that works is returned.
    If all fail an ValType exception is returned including varname
    '''
    for t in expectedTypes:
        if(t == type(value)):
            return value
        else:
            try:
                return t(value)
            except EOQ_ERROR_INVALID_TYPE:
                pass #do nothing since the next type might work 
    #if going here all cast failed:
    expectedTypeStrs = " or ".join([t.__name__ for t in expectedTypes])
    raise EOQ_ERROR_INVALID_TYPE("%s: expected %s but got %s."%(varname,expectedTypeStrs,type(value).__name__))

def ValCompare(a:Any,b:Any)->bool:
    '''Compares a and b for having equal values. 
    It compares values based on primitive values not based on data types, e.g. 
    U32(1) and 1 will be equal.
    For lists it does a deep compare, e.g. 
    LST([]) and [] will be equal.
    '''
    #In Python this is easy, because comparision by value is already implemented by the VAL classes.
    return a==b;


    
### VALUE FACTORY ###
    
def ValFactory(valType:str, args:list)->VAL:
    try: 
        valInfo = VAL_TYPE_REGISTRY[valType]
        if(valInfo.unpackArgs):
            return valInfo.clazz(*args) #create a new instance
        else:
            return valInfo.clazz(args) #create a new instance
    except Exception:#KeyError:
        raise EOQ_ERROR_INVALID_TYPE('Value type %s does not exist.'%(valType))






'''
 Bjoern Annighoefer 2024
'''

from ...domain import Domain
from ...value import STR
from ...query import Obj, Qry
from ...command import Get
from ...concepts import CONCEPTS, MXELEMENT
from ...error import EOQ_ERROR_INVALID_VALUE

from .cmdstream import FromDomainCmdStream, ResNameGenStream, EoqFileOutStream, DependencyResolverStream, ObjToHisStream
from .resnamegen import FromDomainResNameGen

def SaveEoqFile(outfile:str, rootObj:Obj, domain:Domain, sessionId:str=None):
    '''Save a model to an eoqfile
    '''
    #build stream pipeline
    inStream = FromDomainCmdStream(domain,sessionId)
    sorter = DependencyResolverStream(True,False)
    renamer = ResNameGenStream(FromDomainResNameGen(domain,sessionId))
    tohist = ObjToHisStream()
    outStream = EoqFileOutStream(outfile)
    inStream.Connect(sorter)
    sorter.Connect(renamer)
    renamer.Connect(tohist)
    tohist.Connect(outStream)
    #run stream pipeline
    inStream.Begin()
    inStream.LoadElement(rootObj)
    inStream.Flush()
    


'''
2019 Bjoern Annighoefer
'''

from ..error import EOQ_ERROR_INVALID_VALUE
from ..config import Config
# external imports
import traceback
# type annotations
from typing import List,Callable,Type,Dict

class LOG_LEVEL_NAMES:
    ERROR =  "error"
    WARN =   "warning"
    INFO =   "info"
    DEBUG =  "debug"

class LOG_LEVELS_NUM:
    ERROR =  int(1)
    WARN =   int(2)
    INFO =   int(4)
    DEBUG =  int(8)

LOG_LEVELS_NAME_LUT:Dict[int,str] = {
    LOG_LEVELS_NUM.ERROR: LOG_LEVEL_NAMES.ERROR,
    LOG_LEVELS_NUM.WARN: LOG_LEVEL_NAMES.WARN,
    LOG_LEVELS_NUM.INFO: LOG_LEVEL_NAMES.INFO,
    LOG_LEVELS_NUM.DEBUG: LOG_LEVEL_NAMES.DEBUG,
}

# class DEFAULT_LOGGER_LEVELS:
#     L0_SILENT  = []
#     L1_ERROR   = [LOG_LEVEL_NAMES.ERROR]
#     L2_WARNING = [LOG_LEVEL_NAMES.WARN, LOG_LEVEL_NAMES.ERROR]
#     L3_INFO    = [LOG_LEVEL_NAMES.INFO, LOG_LEVEL_NAMES.WARN, LOG_LEVEL_NAMES.ERROR]
#     L4_DEBUG   = [LOG_LEVEL_NAMES.INFO, LOG_LEVEL_NAMES.WARN, LOG_LEVEL_NAMES.ERROR, LOG_LEVEL_NAMES.DEBUG]

class DEFAULT_LOGGER_LEVELS:
    L0_SILENT  = int(0)
    L1_ERROR   = LOG_LEVELS_NUM.ERROR
    L2_WARNING = LOG_LEVELS_NUM.ERROR | LOG_LEVELS_NUM.WARN
    L3_INFO    = LOG_LEVELS_NUM.ERROR | LOG_LEVELS_NUM.WARN | LOG_LEVELS_NUM.INFO
    L4_DEBUG   = LOG_LEVELS_NUM.ERROR | LOG_LEVELS_NUM.WARN | LOG_LEVELS_NUM.INFO | LOG_LEVELS_NUM.DEBUG

class Logger:
    """Base class for all loggers.
    Provides the interface for logging specific log levels, i.e. Debug, Info, Warn, Error.
    Provides the interface for passivatable logging, i.e. PDebug, PInfo, PWarn, PError.
    Each inheriting logger implementation must implement the _Log method.
    """
    def __init__(self, activeLevels:int=DEFAULT_LOGGER_LEVELS.L2_WARNING):
        self.activeLevels = activeLevels
        
    def ShallLog(self):
        return True
    
    def Log(self, level:int, msg:str):
        if(level & self.activeLevels and self.ShallLog()):
            self._Log(LOG_LEVELS_NAME_LUT[level],msg)
    
    #simple log functions
    
    def Debug(self, msg:str):
        self.Log(LOG_LEVELS_NUM.DEBUG, msg)
    
    def Info(self, msg:str):
        self.Log(LOG_LEVELS_NUM.INFO, msg)
        
    def Warn(self, msg:str):
        self.Log(LOG_LEVELS_NUM.WARN, msg)
        
    def Error(self, msg:str):
        self.Log(LOG_LEVELS_NUM.ERROR, msg)
        
    #passive log functions
    
    def PLog(self, level:int, msgFactory:Callable[[],str]):
        """Passivatable log. Is supposed to have higher waste less
        performance if not used, because msgFactory is only evaluated if
        log level is active"""
        if(level & self.activeLevels and self.ShallLog()):
            try:
                levelName = LOG_LEVELS_NAME_LUT[level]
                self._Log(levelName,msgFactory())
            except Exception as e:
                traceback.print_exc()
                self._Log(LOG_LEVEL_NAMES.ERROR,'ERROR while logging: %s'%str(e))
                
    def PDebug(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVELS_NUM.DEBUG, msgFactory)
        
    def PInfo(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVELS_NUM.INFO, msgFactory)
        
    def PWarn(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVELS_NUM.WARN, msgFactory)
    
    def PError(self,msgFactory:Callable[[],str]):
        self.PLog(LOG_LEVELS_NUM.ERROR, msgFactory)
    

    #the following must be overwritten to produce the output
    def _Log(self, levelName:str, msg:str):
        raise NotImplementedError('_Log must be implemented by inheriting class.')

### LOGGER IMPLEMENTATIONS ###
class LoggerInfo:
    def __init__(self, name:str, loggerClass:Type[Logger], initArgNames:List[str]):
        self.name = name
        self.loggerClass = loggerClass
        self.initArgNames = initArgNames

LOGGER_REGISTRY:Dict[str,LoggerInfo] = {} # a registry for for current logger implementations

def RegisterLogger(name:str, loggerClass:Type[Logger], initArgNames:List[str]):
    LOGGER_REGISTRY[name] = LoggerInfo(name, loggerClass, initArgNames)

### LOGGER INSTANCES ###
# Keep registry for logger instances, to reduce resource usage
LOGGER_INSTANCE_REGISTRY:Dict[str,Logger] = {} # a registry for for current logger instances

def LoggerIdString(config:Config)->str:
    '''Returns a string identifying the logger instance
    '''
    loggerIdStr = '%s%s%s'%(config.logger,config.activeLogLevels,config.loggerInitArgs)
    return loggerIdStr

def GetLoggerInstance(config:Config)->Logger:
    '''Returns a logger instance based on the config settings
    '''
    loggerIdStr = LoggerIdString(config)
    try:
        return LOGGER_INSTANCE_REGISTRY[loggerIdStr]
    except KeyError:
        #create a new logger instance
        loggerInfo = LOGGER_REGISTRY[config.logger]
        if(len(loggerInfo.initArgNames) < len(config.loggerInitArgs)):
            raise EOQ_ERROR_INVALID_VALUE('Expecting only %d arguments for logger %s, but got %d'%(len(loggerInfo.initArgNames),config.logger,len(config.loggerInitArgs)))
        for argName in loggerInfo.initArgNames:
            if(not argName in config.loggerInitArgs):
                raise EOQ_ERROR_INVALID_VALUE('Logger argument %s is missing in config'%(argName))
        loggerInstance = loggerInfo.loggerClass(config.activeLogLevels,**config.loggerInitArgs)
        LOGGER_INSTANCE_REGISTRY[loggerIdStr] = loggerInstance
        return loggerInstance

        


'''
 Bjoern Annighoefer 2024
'''

from ...domain import Domain
from ...value import VAL
from ...value import VAL, QRY, STR
from ...query import Qry, IsObj
from ...command import Get
from ...concepts import *
from ...error import EOQ_ERROR_INVALID_TYPE

from typing import Dict, Any

class ResNameGenA():
    ''' Interface to generate unique names for resNames
    '''
    def __init__(self):
        raise NotImplementedError()
        
    def GenResName(self, res:VAL)->str:
        ''' Returns a unique name for a set of results, usually Obj.
        For the same res the same name will be recovered. 
        '''
        raise NotImplementedError()
    
    def InventResName(self)->str:
        ''' Returns a unique name generated out of nothing, but is unique
        '''
        raise NotImplementedError()
    
    
    
class SimpleResNameGen(ResNameGenA):
    ''' Helper to generate readable names for resNames by consecutively enumerating elements.
    '''
    def __init__(self, prefix:str="o"):
        self.prefix = prefix
        self.objCount = 0
        self.objNameLut:Dict[VAL,str] = {}
        
    #@Override
    def GenResName(self, res:VAL)->str:
        resName = None
        if(res in self.objNameLut):
            resName = self.objNameLut[res]
        else:
            resName = self.InventResName()
        self.objNameLut[res] = resName
        return resName
    
    #@Override
    def InventResName(self)->str:
        ''' Returns a unique name generated out of nothing, but is unique
        '''
        resName = "%s%d"%(self.prefix,self.objCount)
        self.objCount += 1
        return resName
    
    


from typing import Hashable, Dict, Any

class FromDomainResNameGen(SimpleResNameGen):
    '''Tries to generate readable names for resNames
    based on the concept type
    '''
    def __init__(self, domain:Domain, sessionId:str, prefix:str="o", segmentSeperator="_"):
        super().__init__(prefix)
        self.domain = domain
        self.sessionId = sessionId
        self.segmentSeperator = segmentSeperator
        #internals
        self.usedNames:Dict[str,bool] = {}
        #concept handler table
        self.resNameHandlers = {}
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2ENUM)] = self.__M2EnumResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2OPTIONOFENUM)] = self.__M2OptionOfEnumResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2MODEL)] = self.__M2ModelResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2PACKAGE)] = self.__M2ModelResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2CLASS)] = self.__M2ClassResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2ATTRIBUTE)] = self.__M2AttributeResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2ASSOCIATION)] = self.__M2AssociationResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2COMPOSITION)] = self.__M2CompositionResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M2INHERITANCE)] = self.__M2InheritanceResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M1MODEL)] = self.__M1ModelResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M1OBJECT)] = self.__M1ObjectResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M1ATTRIBUTE)] = self.__M1AttributeResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M1ASSOCIATION)] = self.__M1AssociationResName
        self.resNameHandlers[GetConceptKeyString(CONCEPTS.M1COMPOSITION)] = self.__M1CompositionResName
        
    #@Override
    def GenResName(self, res:VAL)->str:
        resName = None
        if(IsObj(res)):
            conceptId = self.domain.Do( Get( Qry(res).Pth(STR(MXELEMENT.CONCEPT))),self.sessionId).GetVal()
            try:
                conceptKey = GetConceptKeyString(conceptId)
                handler = self.resNameHandlers[conceptKey]
                name = handler(res)
                resNameRaw = self.segmentSeperator.join([conceptKey.lower(),name if name else ""])
                resName = resNameRaw
                suffix = 1
                while(resName in self.usedNames):
                    suffix += 1
                    resName = "%s%d"%(resNameRaw,suffix) #append count to make name unique
                self.usedNames[resName] = True
                self.objNameLut[res] = resName
            except KeyError:
                resName = super().GenResName(res)
        else:
            raise EOQ_ERROR_INVALID_TYPE("Can only generate resNames from res of type Obj.")
        return resName
        
    ### INTERNALS
        
    def __M2EnumResName(self, target:QRY)->str:
        return self.domain.Do( Get( Qry(target).Pth(STR(M2ENUM.NAME))),self.sessionId).GetVal()
    
    def __M2OptionOfEnumResName(self, target:QRY)->str:
        enumName = self.domain.Do( Get( Qry(target).Pth(STR(M2OPTIONOFENUM.ENUM)).Pth(STR(M2ENUM.NAME))),self.sessionId).GetVal()
        name = self.domain.Do( Get( Qry(target).Pth(STR(M2OPTIONOFENUM.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([enumName,name])

    def __M2ModelResName(self, target:QRY)->str:
        return self.domain.Do( Get( Qry(target).Pth(STR(M2PACKAGE.NAME))),self.sessionId).GetVal()
    
    def __M2PackageResName(self, target:QRY)->str:
        return self.domain.Do( Get( Qry(target).Pth(STR(M2PACKAGE.NAME))),self.sessionId).GetVal()
    
    def __M2ClassResName(self, target:QRY)->str:
        return self.domain.Do( Get( Qry(target).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
    
    def __M2AttributeResName(self, target:QRY)->str:
        className = self.domain.Do( Get( Qry(target).Pth(STR(M2ATTRIBUTE.CLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        name = self.domain.Do( Get( Qry(target).Pth(STR(M2ATTRIBUTE.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([className,name])
    
    def __M2AssociationResName(self, target:QRY)->str:
        className = self.domain.Do( Get( Qry(target).Pth(STR(M2ASSOCIATION.SRCCLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        name = self.domain.Do( Get( Qry(target).Pth(STR(M2ASSOCIATION.SRCNAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([className,name])
    
    def __M2CompositionResName(self, target:QRY)->str:
        className = self.domain.Do( Get( Qry(target).Pth(STR(M2COMPOSITION.PARENTCLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        name = self.domain.Do( Get( Qry(target).Pth(STR(M2COMPOSITION.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([className,name])
    
    def __M2InheritanceResName(self, target:QRY)->str:
        subClassName = self.domain.Do( Get( Qry(target).Pth(STR(M2INHERITANCE.SUBCLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        superClassName = self.domain.Do( Get( Qry(target).Pth(STR(M2INHERITANCE.SUPERCLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([subClassName,superClassName])
    
    def __M1ModelResName(self, target:QRY)->str:
        modelName = self.domain.Do( Get( Qry(target).Pth(STR(M1MODEL.M2MODEL)).Pth(STR(M2PACKAGE.NAME))),self.sessionId).GetVal()
        name = self.domain.Do( Get( Qry(target).Pth(STR(M1MODEL.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([modelName,name])
    
    def __M1ObjectResName(self, target:QRY)->str:
        return super().GenResName(target)
        # className = self.domain.Do( Get( Qry(target).Pth(STR(M1OBJECT.M2CLASS)).Pth(STR(M2CLASS.NAME))),self.sessionId).GetVal()
        # name = self.domain.Do( Get( Qry(target).Pth(STR(M1OBJECT.NAME))),self.sessionId).GetVal()
        # return self.segmentSeperator.join([className,name])
    
    def __M1AttributeResName(self, target:QRY)->str:
        parentName = self.objNameLut[self.domain.Do( Get( Qry(target).Pth(STR(M1ATTRIBUTE.OBJECT))),self.sessionId)]
        className = self.domain.Do( Get( Qry(target).Pth(STR(M1ATTRIBUTE.M2ATTRIBUTE)).Pth(STR(M2ATTRIBUTE.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([parentName[CONCEPT_UNIQUE_LEN+1:],className]) #CONCEPT_UNIQUE_LEN+1 removes the previous prefix
    
    def __M1AssociationResName(self, target:QRY)->str:
        parentName = self.objNameLut[self.domain.Do( Get( Qry(target).Pth(STR(M1ASSOCIATION.SRC))),self.sessionId)]
        className = self.domain.Do( Get( Qry(target).Pth(STR(M1ASSOCIATION.M2ASSOCIATION)).Pth(STR(M2ASSOCIATION.SRCNAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([parentName[CONCEPT_UNIQUE_LEN+1:],className]) #CONCEPT_UNIQUE_LEN+1 removes the previous prefix
    
    def __M1CompositionResName(self, target:QRY)->str:
        parentName = self.objNameLut[self.domain.Do( Get( Qry(target).Pth(STR(M1COMPOSITION.PARENT))),self.sessionId)]
        className = self.domain.Do( Get( Qry(target).Pth(STR(M1COMPOSITION.M2COMPOSITION)).Pth(STR(M2COMPOSITION.NAME))),self.sessionId).GetVal()
        return self.segmentSeperator.join([parentName[CONCEPT_UNIQUE_LEN+1:],className]) #CONCEPT_UNIQUE_LEN+1 removes the previous prefix